library geiger_localstorage;

import 'dart:convert';
import 'dart:typed_data';

import '../../geiger_localstorage.dart';

/// <p>Helper class for serialization serializes important java primitives.</p>
class SerializerHelper {
  static const int stringUid = 123798371293;
  static const int longUid = 1221312393;
  static const int intUid = 122134568793;
  static const int stacktraceUid = 9012350123956;

  static const int intSize = 4;
  static const int longSize = 8;

  static void writeRawInt(ByteSink sink, int value) async {
    sink.sink.add(intToByteArray(value));
  }

  static Future<int> readRawInt(ByteStream stream) async {
    return byteArrayToInt(await stream.popArray(intSize));
  }

  static void writeRawLong(ByteSink sink, int value) async {
    sink.sink.add(longToByteArray(value));
  }

  static Future<int> readRawLong(ByteStream stream) async {
    return byteArrayToLong(await stream.popArray(longSize));
  }

  static Future<List<int>> peekRawLong(ByteStream in_, int num) async {
    ByteStream bs = ByteStream(null, await in_.peekBytes(longSize * num));
    List<int> ret = [];
    for (int i = 0; i < num; i++) {
      ret.add(await readRawLong(bs));
    }
    return ret;
  }

  /// Convert [bytes] to a signed 64bit int.
  static int byteArrayToInt(List<int> bytes) {
    final data = ByteData(4);
    data.buffer.asUint8List().setAll(0, bytes);
    return data.getInt32(0);
  }

  /// Convert signed 32bit int [value] to bytearray of length [size].
  static List<int> intToByteArray(int value) {
    return (ByteData(4)..setInt32(0, value)).buffer.asUint8List();
  }

  /// Convert [bytes] to a signed 64bit int.
  static int byteArrayToLong(List<int> bytes) {
    final data = ByteData(8);
    data.buffer.asUint8List().setAll(0, bytes);
    return data.getInt64(0);
  }

  /// Convert signed 64bit int [value] to bytearray of length [size].
  static List<int> longToByteArray(int value) {
    return (ByteData(8)..setInt64(0, value)).buffer.asUint8List();
  }

  /// <p>Serialize a long variable.</p>
  /// @param out the stream to be read
  /// @param l   the value to be deserialized
  /// @throws IOException if an exception occurs while writing to the stream
  static void writeLong(ByteSink out, int l) {
    writeRawLong(out, longUid);
    writeRawLong(out, l);
  }

  /// <p>Deserialize a long variable.</p>
  /// @param in the stream to be read
  /// @return the deserialized long value
  /// @throws IOException if an exception occurs while writing to the stream
  static Future<int> readLong(ByteStream in_) async {
    castTest('long', longUid, await readRawLong(in_));
    return readRawLong(in_);
  }

  static Future<int> peekLong(ByteStream in_) async {
    List<int> longs = await peekRawLong(in_, 2);
    castTest('long', longUid, longs[1]);
    return longs[2];
  }

  /// <p>Serialize an int variable.</p>
  /// @param out the stream to be read
  /// @param i   the value to be deserialized
  /// @throws IOException if an exception occurs while writing to the stream
  static void writeInt(ByteSink out, int i) {
    writeRawLong(out, intUid);
    writeRawInt(out, i);
  }

  /// <p>Deserialize an int variable.</p>
  /// @param in the stream to be read
  /// @return the deserialized integer value
  /// @throws IOException if an exception occurs while writing to the stream
  static Future<int> readInt(ByteStream in_) async {
    castTest('int', intUid, await readRawLong(in_));
    return readRawInt(in_);
  }

  /// <p>Serialize a string variable.</p>
  /// @param out the stream to be read
  /// @param s   the value to be deserialized
  /// @throws IOException if an exception occurs while writing to the stream
  static void writeString(ByteSink out, String? s) {
    SerializerHelper.writeRawLong(out, stringUid);
    if (s == null) {
      writeRawInt(out, -1);
    } else {
      List<int> barr = utf8.encode(s);
      writeRawInt(out, barr.length);
      out.sink.add(barr);
    }
  }

  /// <p>Deserialize a string variable.</p>
  /// @param in the stream to be read
  /// @return the deserialized string
  /// @throws IOException if an exception occurs while writing to the stream
  static Future<String?> readString(ByteStream in_) async {
    castTest('String', stringUid, await readRawLong(in_));
    int length = await readRawInt(in_);
    if (length == (-1)) {
      return null;
    } else {
      List<int> arr = await in_.popArray(length);
      return utf8.decode(arr);
    }
  }

  /// <p>Serialize an array of StackTraces.</p>
  /// @param out the stream to be read
  /// @param ste the value to be deserialized
  /// @throws IOException if an exception occurs while writing to the stream
  static void writeStackTraces(ByteSink out, StackTrace? ste) {
    writeRawLong(out, stacktraceUid);
    writeString(out, ste?.toString());
  }

  /// <p>Deserialize an array of StackTraceElement variable.</p>
  /// @param in the stream to be read
  /// @return the deserialized array
  /// @throws IOException if an exception occurs while writing to the stream
  static Future<StackTrace?> readStackTraces(ByteStream in_) async {
    castTest('StackTrace', stacktraceUid, await readRawLong(in_));
    final string = await readString(in_);
    if (string == null) return null;
    return StackTrace.fromString(string);
  }

  /// short casting test.
  ///
  /// Verifies if the [expected] result matches the result we [really] got and
  /// throws an exception if they do not matcch.
  static void castTest(String obj, int expected, int really, [int? num]) {
    if (expected != really) {
      String s =
          'cannot cast to ' + obj + ' (expected: $expected; got: $really)';
      if (num != null) {
        s += ' ($num)';
      }
      throw StorageException(s);
    }
  }

  /// Read an object from ByteArrayInputStream.
  /// @param in the byteArrayInputStream to use
  /// @return return the object read
  /// @throws IOException if object cannot be read
/*    static Object readObject(Stream<List<int>> in_)
    {
        switch (("" + SerializerHelper.readIntLong(in_))) {
            case ("" + SerializerHelper.STRING_UID):
                List<int?> arr = [];
                arr = readIntInt(in_);
                in_.read(arr);
                return utf8.encode(arr);
            case ("" + SerializerHelper.LONG_UID):
                return readIntLong(in_);
            default:
                throw new Exception("Cannot cast");
        }
    }*/

  /// Write an object to ByteArrayOutputStream.
  /// @param out the ByteArrayOutputStream to use
  /// @param o the Object to write
  /// @throws IOException if object cannot be written
/*    static void writeObject(Sink<List<int>> out, dynamic o)
    {
        switch (o.runtimeType.toString()) {
            case "String":
                writeString(out, (String().(o)));
                break;
            case "long":
                writeLong(out, o);
                break;
            case "int":
                writeInt(out, o);
                break;
            default:
                throw new Exception("Cannot cast ${o.runtimeType.toString()}");
        }
    }*/

}
